/* 
net-3-driver for the Olicom 2335

This is an extension to the Linux operating system, and is covered by the
same Gnu Public License that covers that work.

Copyright 2001 by Alfred Arnold (alfred@ccac.rwth-aachen.de, aarnold@elsa.de)

This driver is based on the IBMLANA driver, which is itself based on the
SK_G16 and 3C523 driver.

paper sources:
  'PC Hardware: Aufbau, Funktionsweise, Programmierung' by 
  Hans-Peter Messmer for the basic Microchannel stuff
  
  'Linux Geraetetreiber' by Allesandro Rubini, Kalle Dalheimer
  for help on Ethernet driver programming

  'DP83800AVF 10/100 Mb/s ISA Ethernet Contoller' datasheet by National
  Semiconductor for info on the MAC chip

  Also see http://www.natsemi.com/

special acknowledgements to:
  - Peter Wendt for providing the necessary hardware (an Olicom 2335 board)

NOTE: The Tx performance of this driver when connected to a 10 MBit network
      is not optimal (~800KB/s) which is due to the small Tx FIFO of the
      DP83800.  This problem does not occur in 100 MBit mode since the FIFO
      never gets full in this mode.  I thought about adding a software FIFO
      in the driver, but it wasn't worth the effort and anyone who spends
      a lot of money on the only 100 MBit MCA adapter ever made is probably
      not going to hang it on a 10 MBit wire 8-)

History:
  Apr 3rd, 2001
  	startup from IBMLANA driver
  Apr 7th, 2001
	receiver part works
  Apr 8th, 2001
	started work on transmitting
  Apr 9th, 2001
        transmitting works, however with suboptimal speed
  Apr 13th, 2001
	implement Tx errors
  Apr 28th, 2001
	also compilable in-kernel
  Jan 9th, 2002
	fixed address computation - seems noone so far has used his
	OC2335 at a different address than 0x200 so far...

 *************************************************************************/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/mca.h>
#include <linux/version.h>
#include <asm/processor.h>
#include <asm/bitops.h>
#include <asm/io.h>

#ifdef MODULE
#include <linux/module.h>
#endif

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#define _OC2335_DRIVER_
#include "oc2335.h"

#undef DEBUG

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

/* ------------------------------------------------------------------------
 * private subfunctions
 * ------------------------------------------------------------------------ */

/* deduce resources out of POS registers */

static void getaddrs(int slot, int *iobase, int *irq)
{
  u_char pos0, pos3;
  static int oc2335_irqs[8] = {3, 4, 5, 9, 10, 11, 12, 15};
  
  pos0 = mca_read_stored_pos(slot, 2);
  pos3 = mca_read_stored_pos(slot, 5);
  
  *iobase = ((pos0 & 0x0e) << 5) + 0x200;
  *irq = oc2335_irqs[(pos3 & 0x70) >> 4];
}

/* wait on register value with mask and timeout */

static int wait_timeout(struct OC2335_NETDEV *dev, int regoffs, u16 mask,
                        u16 value, int timeout)
{
  unsigned long fin = jiffies + timeout;

  while (jiffies != fin)
    if ((inw(dev->base_addr + regoffs) & mask) == value)
      return 1;

  return 0;
}

#ifdef DEBUG
/* dump registers */

static void dumpregs(struct OC2335_NETDEV *dev)
{
  int z;

  printk("Regdump:\n");
  for (z = 0; z < 32; z+= 2)
  {
    printk("%04x ", inw(dev->base_addr + z));
    if ((z & 15) == 14) 
      printk("\n");
  }
}
#endif


/* initialize the board */

static void InitBoard(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;

  /* first, set up the receiver.  For the moment, we want an
     interrupt for every frame received and no early/burst tricks. */

  outw(1, dev->base_addr + REG_CR);
  outw(0, dev->base_addr + REG_RTR);

  /* allow receiption while transmitting, to allow full-duplex operation.
     automatically discard broken packets. */

  outw(0, dev->base_addr + REG_CR);
  outw(REG_RXCR_ATX, dev->base_addr + REG_RXCR);

  /* next, set up transmitter.  We want an ack for errored frames, do not
     want to extend too short frames, and allow full-duplex mode. */

  outw(0, dev->base_addr + REG_CR);
  outw(REG_TXCR_HBI | REG_TXCR_ATP | REG_TXCR_HERR, dev->base_addr + REG_TXCR);

  /* normal transmit mode: all data must be in the FIFO before transmission
     starts. */

  outw(1, dev->base_addr + REG_CR);
  outw(PKTSIZE, dev->base_addr + REG_TMTR);

  /* generate interrupt if there is space for at least one maximum-length
     frame.  For reasons I do not understand, the NIC interrupts substantially
     before the requested number is actually received. */

  outw(RND_PKTSIZE + 32, dev->base_addr + REG_TFTR);

  /* enable transmitter + receiver */

  outw(REG_CR_MOD | REG_CR_TXE | REG_CR_RXE, dev->base_addr + REG_CR);

  /* enable allowed interrupts, afterwards set general interrupt enable. */

  outw(0, dev->base_addr + REG_CR);
  outw(REG_IMR_RXC | REG_IMR_TXB | REG_IMR_TFS, dev->base_addr + REG_IMR);  

  outw(1, dev->base_addr + REG_CR);
  outw(REG_IER_IE, dev->base_addr + REG_IER);

  /* we start with Tx handle 0 */

  priv->txhandle = 0;
}

/* reset the whole board */

static void ResetBoard(struct OC2335_NETDEV *dev)
{
  /* set reset bits for a while */

  outw(REG_CR_SR | REG_CR_TXR | REG_CR_RXR | REG_CR_MOD, 
       dev->base_addr + REG_CR);
}

/* set up a CAM entry.  Assumes register bank 2 is selected. CAM is 
   disabled afterwards and has to be reenabled ! */

static void setcam(struct OC2335_NETDEV *dev, int *count, char *saddr)
{
  u8 *addr = (u8*) saddr;
  int z;

  /* disable CAM, start write operation.  Since the CAM has to be enabled
     with a write to CCR afterwards, we do not conserve any other bits. */

  outw(REG_CCR_CMW | ((*count) & 0x0f), dev->base_addr + REG_CCR);

  /* successively write 3 words for address, starting with LSB. */

  for (z = 0; z < 3; z++)
  {
    outw((((u16) addr[1]) << 8) | addr[0], dev->base_addr + REG_CDR);
    addr += 2;
  }

  /* next write operation goes to next CAM entry. */

  (*count)++;
}

/* transmit a frame */

static void transmit_skb(struct OC2335_NETDEV *dev, struct sk_buff *skb)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  int dcount;

  /* compute expanded size of frame. */

  dcount = (skb->len + 3) >> 2;

  /* no space in FIFO? Then we shouldn't be here ;-) */

  if ((inw(dev->base_addr + REG_TFSR) >> 2) < (dcount + 1))
  {
    OC2335_KFREE(skb);
    priv->stat.tx_dropped++;
    return;
  }

  /* write frame length into FIFO */

  outw(skb->len, dev->base_addr + REG_TRDR0);

  /* write frame handle to FIFO */

  outw(priv->txhandle++, dev->base_addr + REG_TRDR0);

  /* write frame data to FIFO.  Implicitly write trailing pad bytes. */

  outsl(dev->base_addr + REG_TRDR0, skb->data, dcount);

  /* bookkeeping */

  priv->stat.tx_packets++;
#if (LINUX_VERSION_CODE >= 0x020119)
  priv->stat.tx_bytes += skb->len;
#endif

  /* don't forget 8-) */

  OC2335_KFREE(skb);
}

/* read/write from device Tx FIFO.
   This is an unfinished idea to add a software Tx FIFO in the driver
   to compensate for the small hardware Tx FIFO.  Turned out however
   that the win was near to zero... */

#if 0
static struct sk_buff *getskb(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  struct sk_buff *skb;

  if (!priv->skbcnt)
    return NULL;

  skb = priv->skbs[0];
  memmove(priv->skbs, priv->skbs + 1, sizeof(struct sk_buff *) * (TXBUFCNT - 1));
  priv->skbcnt--;

  return skb;
}

static void putskb(struct OC2335_NETDEV *dev, struct sk_buff *skb)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;

  /* FIFO full ?  Then dump the oldest frame */

  if (priv->skbcnt >= TXBUFCNT)
    OC2335_KFREE(getskb(dev));

  /* add at end of FIFO */

  priv->skbs[priv->skbcnt++] = skb;
}
#endif

/* ------------------------------------------------------------------------
 * interrupt handler(s)
 * ------------------------------------------------------------------------ */

/* receive interrupt */

static void irqrx_handler(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  struct sk_buff *skb;
  int bcount, dcount;
  u16 status;

  /* get frame length - omit 4 bytes of trailing CRC */

  bcount = (inw(dev->base_addr + REG_RSR) & REG_RSR_BC) - 4;
  dcount = (bcount + 3) >> 2;

  /* get memory for frame */

  skb = dev_alloc_skb((dcount << 2) + 2);

  if (skb == NULL)
    priv->stat.rx_dropped++;
  else
  {
    /* copy out data */

    insl(dev->base_addr + REG_TRDR0, skb_put(skb, dcount << 2), dcount);

    /* adjust length */

    skb->len -= (dcount << 2) - bcount;

    /* set up skb fields */

    skb->dev = dev;
    skb->protocol = eth_type_trans(skb, dev);
    skb->ip_summed = CHECKSUM_NONE;

    /* bookkeeping */

    priv->stat.rx_packets++;
#if (LINUX_VERSION_CODE >= 0x20119)  /* byte counters for kernel >= 2.1.25 */
    priv->stat.rx_bytes += bcount;
#endif

    /* pass to the upper layers */

    netif_rx(skb);
  }

  /* get status, thereby acknowledge packet */

  status = inw(dev->base_addr + REG_RPAR);
}

/* interrupt signaling there is again enough space in FIFO */

static void irqtxfree_handler(struct OC2335_NETDEV *dev)
{
  u16 freeval;

  /* check if there is really enough space: */

  freeval = inw(dev->base_addr + REG_TFSR);
  if (freeval >= RND_PKTSIZE)
  {
    /* yes: wake up the transmitter */

#if LINUX_VERSION_CODE >= 0x02032a
    netif_wake_queue(dev);
#else
    dev->tbusy = 0;
    mark_bh(NET_BH);
#endif
  }
  else
  {
    /* no: should never occur... */

    printk("%s: signaled enough space but not enough available (%d<%d).\n",
           dev->name, freeval, inw(dev->base_addr + REG_TFTR));
    ResetBoard(dev);
  }
}

/* interrupt signaling tx error */

static void irqtxerr_handler(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  u16 errval;

  /* read error code */

  errval = inw(dev->base_addr + REG_TSR);

  /* update statistics */

  priv->stat.tx_errors++;
  if (errval & REG_TSR_CRS)
    priv->stat.tx_carrier_errors++;
  if (errval & (REG_TSR_TD | REG_TSR_ED | REG_TSR_EC))
    priv->stat.tx_aborted_errors++;
  if (errval & REG_TSR_OWC)
    priv->stat.tx_window_errors++;
  if (errval & REG_TSR_TFU)
    priv->stat.tx_fifo_errors++;
}

/* general interrupt entry */

static void irq_handler(int irq, void *device, struct pt_regs *regs)
{
  struct OC2335_NETDEV *dev = (struct OC2335_NETDEV*) device;
  u16 ival;

  /* in case we're not meant... */

  ival = inw(dev->base_addr + REG_IVR);
  if (!ival)
    return;

#if (LINUX_VERSION_CODE >= 0x02032a)
 #if 0
  set_bit(LINK_STATE_RXSEM, &dev->state);
 #endif
#else
  dev->interrupt = 1;
#endif

  /* query interrupt vector register until everything is clear */

  while (ival)
  {
    switch (ival)
    {
      case REG_IVR_RXC:
        irqrx_handler(dev);
        break;
      case REG_IVR_TFS:
        irqtxfree_handler(dev);
        break;
      case REG_IVR_TXB:
        irqtxerr_handler(dev);
        break;
      default:
        printk("%s: unknown interrupt vector %02x, turning board off :-(\n",
               dev->name, ival);
        ResetBoard(dev);
    }

    /* get next condition */

    ival = inw(dev->base_addr + REG_IVR);
  }

#if (LINUX_VERSION_CODE >= 0x02032a)
 #if 0
  clear_bit(LINK_STATE_RXSEM, &dev->state);
 #endif
#else
  dev->interrupt = 0;
#endif
}

/* ------------------------------------------------------------------------
 * driver methods
 * ------------------------------------------------------------------------ */

/* switch receiver mode. */

static void oc2335_set_multicast_list(struct OC2335_NETDEV *dev)
{
  int camcnt = 0;
  struct dev_mc_list *mcptr;
  u16 ctrlval;

  /* select register page 2 for CAM */

  outw(2, dev->base_addr + REG_CR);

  /* we have to disable the CAM before we can access it. this is done
     implicitly with the write to the first CAM entry, which holds the
     card's physical address. */

  setcam(dev, &camcnt, dev->dev_addr);

  /* program other multicast addresses.  We have to stop when the NIC's
     multicast CAM entries are used up and will set the AAM flag later. */

  for (mcptr = dev->mc_list; mcptr != NULL; mcptr = mcptr->next)
  {
    setcam(dev, &camcnt, mcptr->dmi_addr);
    if (camcnt >= 10)
      break;
  }

  /* calculate CAM mask. */

  outw((1 << camcnt) - 1, dev->base_addr + REG_CMR);

#ifdef DEBUG
  printk("CAM Contents:");
  for (camcnt = 0; camcnt < 10; camcnt++)
  {
    int z;

    outw(REG_CCR_CMR | camcnt, dev->base_addr + REG_CCR);
    printk("CAM entry %d: ", camcnt);
    for (z = 0; z < 3; z++)
      printk("%04x ", inw(dev->base_addr + REG_CDR));
    printk("\n");
  }
#endif

  /* build value for control register.  Always accept all broadcasts, all
     multicasts if either requested or too many addresses, and promiscous
     mode if requested. */

  ctrlval = REG_CCR_CMEN | REG_CCR_AAB;
  if ((dev->flags & IFF_ALLMULTI) || (mcptr != NULL))
    ctrlval |= REG_CCR_AAM;
  if (dev->flags & IFF_PROMISC)
    ctrlval |= REG_CCR_AAP;
  outw(ctrlval, dev->base_addr + REG_CCR);

  /* reenable register bank 1 */

  outw(1, dev->base_addr + REG_CR);
}

/* MCA info */

static int oc2335_getinfo(char *buf, int slot, void *d)
{
  int len = 0, i;
  struct OC2335_NETDEV *dev = (struct OC2335_NETDEV*) d;
  oc2335_priv *priv;
  
  /* can't say anything about an uninitialized device... */

  if (dev == NULL)
    return len;
  if (dev->priv == NULL)
    return len;
  priv = (oc2335_priv*) dev->priv;

  /* print info */

  len += sprintf(buf + len, "IRQ: %d\n", priv->realirq);
  len += sprintf(buf + len, "I/O: %#lx\n", dev->base_addr);
  len += sprintf(buf + len, "Device: %s\n", dev->name);
  len += sprintf(buf + len, "MAC address:");
  for (i = 0; i < 6; i ++ )
    len += sprintf( buf+len, " %02x", dev->dev_addr[i] );
  buf[len++] = '\n';
  len += sprintf(buf + len, "Silicon revision: %x\n", priv->revision);
  buf[len] = 0;

  return len;
}

/* open driver. Means also initialization of the NIC. */

static int oc2335_open(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  int z, result;

  /* register resources - only necessary for IRQ */

  result = request_irq(priv->realirq, irq_handler, SA_SHIRQ | SA_SAMPLE_RANDOM,
                       "oc2335", dev);
  if (result != 0)
  {
    printk("%s: failed to register irq %d\n", dev->name, dev->irq);
    return result;
  }
  dev->irq = priv->realirq;

  /* set up the card and NIC */

  InitBoard(dev);

  /* no Tx buffer stored ahead */

  priv->skbcnt = 0;
  for (z = 0; z < TXBUFCNT; priv->skbs[z++] = NULL);

  /* initialize operational flags */

#if (LINUX_VERSION_CODE >= 0x02032a)
  netif_start_queue(dev);
#else
  dev->interrupt = 0;
  dev->tbusy = 0;
  dev->start = 1;
#endif

#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif

  return 0;
}

/* close driver.  Shut down board and free allocated resources */

static int oc2335_close(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  int z;

  /* turn off board */

  ResetBoard(dev);

  /* ahead buffer to release ? */

  for (z = 0; z < priv->skbcnt; z++)
    OC2335_KFREE(priv->skbs[z++]);
  priv->skbcnt = 0;

  /* release resources */

  if (dev->irq != 0)
    free_irq(dev->irq, dev);
  dev->irq = 0;

#if (LINUX_VERSION_CODE < 0x02032a)
  dev->tbusy = 1;
  dev->start = 0;
#endif

#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif

  return 0;
}

/* transmit a block. */

static int oc2335_tx(struct sk_buff *skb, struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;
  int dcount;
  unsigned long flags;

  /* if we get called with a NULL descriptor, the Ethernet layer thinks
     our card is stuck an we should reset it.  We'll do this completely: */

  if (skb == NULL)
  {
    printk("%s: Resetting NIC\n", dev->name);
    ResetBoard(dev);
    oc2335_set_multicast_list(dev);
    InitBoard(dev);
    return 0; /* don't try to free the block here ;-) */
  }

  /* compute expanded size of frame. */

  dcount = (skb->len + 3) >> 2;

  /* no space in FIFO? Then we shouldn't be here ;-) */

  if ((inw(dev->base_addr + REG_TFSR) >> 2) < (dcount + 1))
  {
    OC2335_KFREE(skb);
    priv->stat.tx_dropped++;
    return -EIO;
  }

  /* transmit frame */

  transmit_skb(dev, skb);

  /* if there not enough space for one full-length frame, disable 
     further transmission */

  save_flags(flags);
  cli();

  if (inw(dev->base_addr + REG_TFSR) < RND_PKTSIZE)
  {
#if (LINUX_VERSION_CODE >= 0x02032a)
    netif_stop_queue(dev);
#else
    dev->tbusy = 1;
#endif
  }

  restore_flags(flags);

  return 0;
}

/* return pointer to Ethernet statistics */

static struct OC2335_STAT *oc2335_stats(struct OC2335_NETDEV *dev)
{
  oc2335_priv *priv = (oc2335_priv*) dev->priv;

  return &(priv->stat);
}

/* we don't support runtime reconfiguration, since am MCA card can
   be unambigously identified by its POS registers. */

static int oc2335_config(struct OC2335_NETDEV *dev, struct ifmap *map)
{
  return 0;
}

/* ------------------------------------------------------------------------
 * hardware check
 * ------------------------------------------------------------------------ */

static int startslot; /* counts through slots when probing multiple devices */

int oc2335_probe(struct OC2335_NETDEV *dev)
{
  int force_detect = 0;
  int slot, z;
  u16 val;
  int base = 0, irq = 0, iobase = 0;
  oc2335_priv *priv;

  /* can't work without an MCA bus ;-) */
  
  if (MCA_bus == 0)
    return ENODEV;
    
  /* start address of 1 --> forced detection */
  
  if (dev->mem_start == 1)
    force_detect = 1;

  /* search through slots */
  
  if (dev != NULL)
  {
    base = dev->mem_start;
    irq = dev->irq;
  }
  slot = mca_find_adapter(OC2335_ID, startslot);

  while (slot != -1)
  {
    /* deduce card addresses */

    getaddrs(slot, &iobase, &irq);

#if (LINUX_VERSION_CODE >= 0x20300)
    /* slot already in use ? */

    if (mca_is_adapter_used(slot))
    {
      slot = mca_find_adapter(OC2335_ID, slot + 1);
      continue;
    }
#endif

    /* were we looking for something different ? */
    
    if ((dev->irq != 0) && (dev->irq != irq))
    {
      slot = mca_find_adapter(OC2335_ID, slot + 1);
      continue;
    }
  
    /* found something that matches */
    
    break;
  }

  /* nothing found ? */
  
  if (slot == -1)
    return ((base != 0) || (irq != 0)) ? ENXIO : ENODEV;
    
  /* announce success */
  printk("%s: Olicom 2335 found in slot %d\n", dev->name, slot + 1);

  /* try to obtain I/O range */
  if (check_region(iobase, OC2335_IORANGE) < 0)
  {
    printk("cannot allocate I/O range at %#x!\n", iobase);
    startslot = slot + 1;
    return 0;
  }
  request_region(iobase, OC2335_IORANGE, "oc2335");

  /* make procfs entries */

  mca_set_adapter_name(slot, "Olicom 2335");
  mca_set_adapter_procfn(slot, (MCA_ProcFn) oc2335_getinfo, dev);

#if (LINUX_VERSION_CODE >= 0x20200)
  mca_mark_as_used(slot);
#endif
  
  /* allocate structure */

  priv = dev->priv = (oc2335_priv*) kmalloc(sizeof(oc2335_priv), GFP_KERNEL);
  priv->slot = slot;
  priv->realirq = irq;
  memset(&(priv->stat), 0, sizeof(priv->stat));
  
  /* set base + irq for this device (irq not allocated so far) */

  dev->irq = 0;
  dev->base_addr = iobase;
 
  /* set methods */

  dev->open = oc2335_open;
  dev->stop = oc2335_close;
  dev->set_config = oc2335_config;
  dev->hard_start_xmit = oc2335_tx;
  dev->do_ioctl = NULL;
  dev->get_stats = oc2335_stats;
  dev->set_multicast_list = oc2335_set_multicast_list;
  dev->flags |= IFF_MULTICAST;

  /* generic setup */

  ether_setup(dev);

  /* reset board */

  ResetBoard(dev);

  /* read silicon revision */

  outw(4, dev->base_addr + REG_CR);
  priv->revision = inw(dev->base_addr + REG_SSR);

  /* copy out MAC address: switch to register page 2 */

  outw(2, dev->base_addr + REG_CR);
  for (z = 0; z < 3; z++)
  {
    wait_timeout(dev, REG_EECR, REG_EECR_EIU, 0, 2);
    outw(REG_EECR_RD | (MACADDROFFS + z), dev->base_addr + REG_EECR);
    wait_timeout(dev, REG_EECR, REG_EECR_EIU, 0, 2);
    val = inw(dev->base_addr + REG_EEDR);
    dev->dev_addr[(z << 1)    ] = val        & 0xff;
    dev->dev_addr[(z << 1) + 1] = (val >> 8) & 0xff;
  }

  /* print config */

  printk("%s: IRQ %d, I/O %#lx, "
         "MAC address %02x:%02x:%02x:%02x:%02x:%02x, silicon revision %x.\n",
         dev->name, priv->realirq,  dev->base_addr,
         dev->dev_addr[0], dev->dev_addr[1], dev->dev_addr[2],
         dev->dev_addr[3], dev->dev_addr[4], dev->dev_addr[5],
         priv->revision);

  /* next probe will start at next slot */

  startslot = slot + 1;

  return 0;
}

/* ------------------------------------------------------------------------
 * modularization support
 * ------------------------------------------------------------------------ */

#ifdef MODULE

#define DEVMAX 5

#if (LINUX_VERSION_CODE >= 0x020363)
static struct OC2335_NETDEV moddevs[DEVMAX] =
       {{"    ", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {"    ", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {"    ", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {"    ", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {"    ", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe}};
#else
static char NameSpace[8 * DEVMAX];
static struct OC2335_NETDEV moddevs[DEVMAX] =
       {{NameSpace +  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {NameSpace +  8, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {NameSpace + 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {NameSpace + 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe},
        {NameSpace + 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, oc2335_probe}};
#endif

int irq = 0;
int io = 0;

int init_module(void)
{
  int z, res;

  startslot = 0;
  for (z = 0; z < DEVMAX; z++)
  {
    strcpy(moddevs[z].name, "     ");
    res = register_netdev(moddevs + z);
    if (res != 0)
      return (z > 0) ? 0 : -EIO;
  }

  return 0;
}

void cleanup_module(void)
{
  struct OC2335_NETDEV *dev;
  oc2335_priv *priv;
  int z;

  if (MOD_IN_USE)
  {
    printk("cannot unload, module in use\n");
    return;
  }

  for (z = 0; z < DEVMAX; z++)
  {
    dev = moddevs + z;
    if (dev->priv != NULL)
    {
      priv = (oc2335_priv*) dev->priv;
      if (dev->irq != 0)
        free_irq(dev->irq, dev);
      dev->irq = 0;
      release_region(dev->base_addr, OC2335_IORANGE);
      unregister_netdev(dev);
#if (LINUX_VERSION_CODE >= 0x20200)
      mca_mark_as_unused(priv->slot);
#endif
      mca_set_adapter_name(priv->slot, "");
      mca_set_adapter_procfn(priv->slot, NULL, NULL);
#if (LINUX_VERSION_CODE >= 0x020369)
      kfree(dev->priv);
#else
      kfree_s(dev->priv, sizeof(oc2335_priv));
#endif
      dev->priv = NULL;
    }
  }
}
#endif /* MODULE */
